package cn.dankal.share.util;

import cn.dankal.share.common.IntArrayFactory;
import cn.dankal.share.common.IntArraySumUtil;
import cn.dankal.share.common.MergeSort;
import cn.dankal.share.forkjoin.FindFileTask;
import cn.dankal.share.forkjoin.IntArraySortTask;
import cn.dankal.share.forkjoin.IntArraySumTask;
import cn.hutool.core.util.RandomUtil;
import jdk.nashorn.internal.ir.annotations.Ignore;
import org.junit.jupiter.api.Test;

import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.TimeUnit;

/**
 * Fork/Join 相关测试类
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class ForkJoinTest {

    /**
     * 测试 Fork/Join 实现的 归并排序
     */
    @Test
    public void testMergeSortByForkJoin() {
        int[] generateIntArray = IntArrayFactory.generateIntArray(25);
        System.out.println("排序前:");
        System.out.println(Arrays.toString(generateIntArray));
        System.out.println("========================");
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        IntArraySortTask intArraySortTask = new IntArraySortTask(generateIntArray, 2);
        int[] invokeResult = forkJoinPool.invoke(intArraySortTask);
        System.out.println("排序后:");
        System.out.println(Arrays.toString(invokeResult));
        System.out.println("========================");
    }

    /**
     * 测试 Fork/Join 实现 数组求和
     */
    @Test
    @Ignore
    public void testArraySortByForkJoin() {
        int[] generateIntArray = IntArrayFactory.generateIntArray(50);
        IntArraySumTask intArraySumTask = new IntArraySumTask(10, generateIntArray, 0, generateIntArray.length - 1);
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        Integer invokeResult = forkJoinPool.invoke(intArraySumTask);
        System.out.println(invokeResult);
    }


    /**
     * int 求和
     *
     * @param left  被加数
     * @param right 加数
     * @return 和
     */
    private static int intSum(int left, int right) {
        try {
            TimeUnit.MILLISECONDS.sleep(100);
        } catch (InterruptedException e) {
            // ignore
        }
        return left + right;
    }

    /**
     * 测试 归并排序对比
     */
    @Test
    public void testMergeSortThreeWays() {
        int[] generateIntArray = IntArrayFactory.generateIntArray(3000000);
        System.out.println("数组元素:  " + generateIntArray.length);

        // general merge sort
        final long generalMergeSortStart = System.currentTimeMillis();
        MergeSort.sort(generateIntArray);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - generalMergeSortStart) + " 毫秒 (general mergeSort)");

        // jdk utils: Arrays#sort
        final long arraySortStartMs = System.currentTimeMillis();
        // DualPivotQuicksort 快排
        Arrays.sort(generateIntArray);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - arraySortStartMs) + " 毫秒 (Arrays#sort)");

        // Fork/Join merge sort
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        IntArraySortTask intArraySortTask = new IntArraySortTask(generateIntArray, 300);
        final long forkJoinStartMs = System.currentTimeMillis();
        forkJoinPool.invoke(intArraySortTask);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - forkJoinStartMs) + " 毫秒 (Fork/Join mergeSort)");

    }

    /**
     * 测试: Fork/Join的优势
     */
    @Test
    public void testAdvantageOfForkJoin() {
        Map<String, Long> expenditureTimeMap = new LinkedHashMap<>(1 << 4);
        final String streamDataKey = "java8 stream";
        final String forkJoinDataKey = "fork join";
        final String generalSumDataKey = "general sum";
        final int count = 10;
        for (int i = 0; i < count; i++) {
            int[] generateIntArray = IntArrayFactory.generateIntArray(RandomUtil.randomInt(50, 100));
            System.out.printf("目标数组 length: %d%n ", generateIntArray.length);
            System.out.println(Arrays.toString(generateIntArray));
            System.out.printf("============= 第 %d 次计算 =============%n", (i + 1));
            final long streamStartMs = System.currentTimeMillis();
            int result = Arrays.stream(generateIntArray)
                    .parallel()
                    .reduce(ForkJoinTest::intSum)
                    .orElse(-1);
            long streamExpenditure = System.currentTimeMillis() - streamStartMs;
            System.out.println("(java8 stream) 运行结果: " + result);
            System.out.println("(java8 stream) 运行耗时: " + streamExpenditure + " 毫秒\n");
            // 将 java8 stream 每次执行时间存储在 LinkedHashMap 中
            if (!expenditureTimeMap.containsKey(streamDataKey)) {
                expenditureTimeMap.put(streamDataKey, streamExpenditure);
            } else {
                Long oldExpenditure = expenditureTimeMap.get(streamDataKey);
                expenditureTimeMap.put(streamDataKey, oldExpenditure + streamExpenditure);
            }

            IntArraySumTask intArraySumTask = new IntArraySumTask(10, generateIntArray, 0, generateIntArray.length - 1);
            ForkJoinPool forkJoinPool = new ForkJoinPool();
            final long forkJoinStartMs = System.currentTimeMillis();
            Integer invokeResult = forkJoinPool.invoke(intArraySumTask);
            long forkJoinExpenditure = System.currentTimeMillis() - forkJoinStartMs;
            System.out.println("(fork/join) 运行结果: " + invokeResult);
            System.out.println("(fork/join) 运行耗时: " + forkJoinExpenditure + " 毫秒\n");
            // 将 fork/join 每次执行时间存储在 LinkedHashMap 中
            if (!expenditureTimeMap.containsKey(forkJoinDataKey)) {
                expenditureTimeMap.put(forkJoinDataKey, forkJoinExpenditure);
            } else {
                Long oldExpenditure = expenditureTimeMap.get(forkJoinDataKey);
                expenditureTimeMap.put(forkJoinDataKey, oldExpenditure + forkJoinExpenditure);
            }

            final long generalStartMs = System.currentTimeMillis();
            int sum = IntArraySumUtil.sumArray(generateIntArray, 0, generateIntArray.length - 1);
            long generalSumExpenditure = System.currentTimeMillis() - generalStartMs;
            System.out.println("(general sum) 运行结果: " + sum);
            System.out.println("(general sum) 运行耗时: " + generalSumExpenditure + " 毫秒");
            // 将 general sum 每次执行时间存储在 LinkedHashMap 中
            if (!expenditureTimeMap.containsKey(generalSumDataKey)) {
                expenditureTimeMap.put(generalSumDataKey, generalSumExpenditure);
            } else {
                Long oldExpenditure = expenditureTimeMap.get(generalSumDataKey);
                expenditureTimeMap.put(generalSumDataKey, oldExpenditure + generalSumExpenditure);
            }
            System.out.println("=============================\n");
        }
        System.out.println("平均耗时");
        for (String key : expenditureTimeMap.keySet()) {
            System.out.printf("%s 运行 %d 次,平均耗时 %d 毫秒%n", key, count, expenditureTimeMap.get(key) / count);
        }

    }

    /**
     * 测试: Fork/Join RecursiveAction: 目标文件查找
     */
    @Test
    public void testFindFileByRecursiveAction() {
        final String targetFolder = "C:/dev/MavenRepository";
        final String findFileSuffix = ".lastUpdated";
        FindFileTask findFileTask = new FindFileTask(targetFolder, findFileSuffix);
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        forkJoinPool.invoke(findFileTask);
        System.out.printf("共遍历 : %d 个文件夹%n", FindFileTask.COUNTER_EACH_FOLDER.get());
        System.out.printf("找到目标文件 %s : %d 个%n", findFileSuffix, FindFileTask.COUNTER_TARGET_FILE.get());
    }
}
