package cn.dankal.share.util.completablefuture;

import cn.dankal.share.common.ThreadExecutorFactory;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

/**
 * <p>
 * <code>CompletableFuture</code> 测试
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-22
 */
public class CompletableFutureTest {

    /*
         需求:
              - 获取商品信息   (A)              耗时 120 毫秒
              - 获取商品价格     (B)            耗时 95 毫秒
              - 商品属性获取并归并(Sku/Spu)  (C)  耗时 165 毫秒
          重要: 我们只有获取到了 '商品信息' 才能执行 '商品属性获取并归并',即 C 必须在 A 之后执行
     */

    /**
     * <code>CompletableFuture</code> 测试
     */
    @Test
    public void test() throws Exception {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("dankal-test-completableFuture");

        // 获取商品信息 耗时:  120 毫秒
        final long getProductElapsed = 120L;
        Supplier<Integer> getProductSupplier = () -> {
            try {
                // do work
                TimeUnit.MILLISECONDS.sleep(getProductElapsed);
            } catch (InterruptedException e) {
                // ignored
            }
            // 返回商品信息成功: 200
            System.out.printf("%s 获取商品信息成功,即将返回%n", Thread.currentThread().getName());
            return 200;
        };

        // 获取商品价格 耗时 95 毫秒
        final long getProductPriceElapsed = 95L;
        Supplier<Double> getProductPriceSupplier = () -> {
            try {
                // do work
                TimeUnit.MILLISECONDS.sleep(getProductPriceElapsed);
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("%s 成功获取到商品价格即将开始返回%n", Thread.currentThread().getName());
            return 135.53;
        };

        // 归并商品属性耗时  165 毫秒
        final long getProductPropertiesElapsed = 165L;
        //程序开始计时
        final long startMs = System.currentTimeMillis();
        // 获取 商品信息 并归并商品属性
        CompletableFuture<Integer> productAboutCompletableFuture = CompletableFuture.supplyAsync(getProductSupplier, threadPoolExecutor)
                .thenApplyAsync(previousResult -> {
                    System.out.printf("%s 获取到商品信息: %d 工耗时 %d 毫秒%n", Thread.currentThread().getName(), previousResult, (System.currentTimeMillis() - startMs));
                    try {
                        // do work
                        TimeUnit.MILLISECONDS.sleep(getProductPropertiesElapsed);
                    } catch (InterruptedException e) {
                        // ignored
                    }
                    System.out.printf("%s 完成商品属性归并,即将返回结果%n", Thread.currentThread().getName());
                    return previousResult * 2;
                }, threadPoolExecutor);
        CompletableFuture<Double> getProductPriceCompletableFuture = CompletableFuture.supplyAsync(getProductPriceSupplier, threadPoolExecutor);

        // 等待程序执行完成
        CompletableFuture.allOf(productAboutCompletableFuture, getProductPriceCompletableFuture).get();
        System.out.println("程序执行完成:");
        System.out.printf("\t运行耗时: %d %n", (System.currentTimeMillis() - startMs));
        System.out.printf("\t理论耗时: %d %n", (getProductElapsed + getProductPropertiesElapsed + getProductPriceElapsed));
        // 模拟 Service 方法返回
        System.out.printf("获取商品详情: %d%n", productAboutCompletableFuture.get()/2);
        System.out.printf("获取商品属性: %d%n", productAboutCompletableFuture.get());
        System.out.printf("获取商品价格: %.2f RMB%n", getProductPriceCompletableFuture.get());
    }
}
